@@ -80,6 +80,9 @@ def vi_bindings(registry, cli_ref):
         """
         Escape goes to vi navigation mode.
         """
+        if event.input_processor.input_mode in (InputMode.INSERT,
+                                                InputMode.VI_REPLACE):
+            line.cursor_position += line.document.get_cursor_left_position()
         if event.input_processor.input_mode == InputMode.SELECTION:
             line.exit_selection()
             event.input_processor.pop_input_mode()
@@ -300,6 +303,7 @@ def vi_bindings(registry, cli_ref):
         Replace single character under cursor
         """
         line.insert_text(event.data * event.arg, overwrite=True)
+        line.cursor_position -= 1
 
     @handle('R', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
